import { beforeEach, describe, expect, it, vitest } from "vitest";
import { SocketType } from "../models/socket";
import { Workflow, WorkflowExecutionContext, WorkflowExecutionResult } from "../models/workflow";
import { BlockExecutionContext, BlockExecutionResult } from "../models/workflow-block";
import { createInputSocket, createOutputSocket } from "../socket.utils";
import { DefaultWorkflowLogger } from "./workflow-logger";
import { WorkflowEngine } from "./workflow.engine";

const MOCK_BLOCK = "mock_block";
const MOCK_BLOCK_CONFIG = "mock_block_config";
const MOCK_BLOCK_DEADEND = "mock_block_deadend";
const MOCK_BLOCK_STUPIDEND = "mock_block_stupidend";
const MOCK_BLOCK_THROWS = "mock_block_throws";

describe("Workflow Engine", () => {
  let workflowEngine: WorkflowEngine;
  let mockBlockActionFn: () => void;

  beforeEach(() => {
    mockBlockActionFn = vitest.fn();
    workflowEngine = new WorkflowEngine();
    workflowEngine.addBlockDefinition(MOCK_BLOCK, {
      getSockets: () => [
        createInputSocket("input", SocketType.NUMBER),
        createOutputSocket("output", SocketType.NUMBER)
      ],
      getConfigurationSchema: () => undefined,
      executeBlock: async (
        executionContext: BlockExecutionContext,
        workflowExecutionContext: WorkflowExecutionContext
      ) => {
        mockBlockActionFn();
        await workflowExecutionContext.logger.info("info");
        await workflowExecutionContext.logger.debug("debug");
        await workflowExecutionContext.logger.error("error");
        await workflowExecutionContext.logger.warn("warn");
        if (typeof executionContext.data === "number") executionContext.data += 1;
        else executionContext.data = 1;
        return {
          socketName: "output",
          data: executionContext.data
        } as BlockExecutionResult;
      }
    });
    workflowEngine.addBlockDefinition(MOCK_BLOCK_DEADEND, {
      getSockets: () => [createInputSocket("input", SocketType.NUMBER)],
      executeBlock: async (
        executionContext: BlockExecutionContext,
        workflowExecutionContext: WorkflowExecutionContext
      ) => {
        //
      },
      getConfigurationSchema: () => undefined
    });
    workflowEngine.addBlockDefinition(MOCK_BLOCK_STUPIDEND, {
      getSockets: () => [createInputSocket("input", SocketType.NUMBER)],
      executeBlock: async (
        executionContext: BlockExecutionContext,
        workflowExecutionContext: WorkflowExecutionContext
      ) => {
        return "huhu" as any;
      },
      getConfigurationSchema: () => undefined
    });
    workflowEngine.addBlockDefinition(MOCK_BLOCK_CONFIG, {
      getSockets: () => [createInputSocket("input", SocketType.NUMBER)],
      executeBlock: async (
        executionContext: BlockExecutionContext,
        workflowExecutionContext: WorkflowExecutionContext
      ) => {
        return { data: executionContext.configuration && executionContext.configuration.meins } as BlockExecutionResult;
      },
      getConfigurationSchema: () => undefined
    });
    workflowEngine.addBlockDefinition(MOCK_BLOCK_CONFIG, {
      getSockets: () => [createInputSocket("input", SocketType.NUMBER)],
      executeBlock: async (
        executionContext: BlockExecutionContext,
        workflowExecutionContext: WorkflowExecutionContext
      ) => {
        return { data: executionContext.configuration && executionContext.configuration.meins } as BlockExecutionResult;
      },
      getConfigurationSchema: () => undefined
    });
    workflowEngine.addBlockDefinition(MOCK_BLOCK_THROWS, {
      getSockets: () => [createInputSocket("input", SocketType.NUMBER)],
      executeBlock: async (
        executionContext: BlockExecutionContext,
        workflowExecutionContext: WorkflowExecutionContext
      ) => {
        throw new Error("test");
      },
      getConfigurationSchema: () => undefined
    });
  });

  it("can trigger simple workflows", async () => {
    const workflow: Workflow = { blocks: [{ id: "blockA", definitionId: MOCK_BLOCK }] };
    await workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" });
    expect(mockBlockActionFn).toHaveBeenCalledTimes(1);
  });

  it("can trigger complex workflows", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: MOCK_BLOCK },
        { id: "blockB", definitionId: MOCK_BLOCK }
      ],
      connections: [
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
      ]
    };
    await workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" });
    expect(mockBlockActionFn).toHaveBeenCalledTimes(2);
  });

  it("can't trigger workflow with wrong starting block", async () => {
    const workflow: Workflow = { blocks: [{ id: "blockA", definitionId: MOCK_BLOCK }] };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "inputsss" })
    ).rejects.toThrowError();
  });

  it("can't trigger a workflow with missing definitionId", async () => {
    const workflow: Workflow = {
      blocks: [{ id: "blockA", definitionId: "fooquark" }],
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" })
    ).rejects.toThrowError("block definition not found by id: fooquark");
  });

  it("can't trigger a workflow with missing start block", async () => {
    const workflow: Workflow = {
      blocks: [{ id: "blockA", definitionId: "fooquark" }],
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockB", { socketName: "input" })
    ).rejects.toThrowError("block not found by id: blockB");
  });

  it("can't connect non-existing sockets", async () => {
    await expect(
      workflowEngine.triggerWorkflowBlock(
        {
          blocks: [
            { id: "blockA", definitionId: MOCK_BLOCK },
            { id: "blockB", definitionId: MOCK_BLOCK }
          ],
          connections: [
            { sourceBlockId: "blockA", sourceSocketName: "foo", targetBlockId: "blockB", targetSocketName: "baa" }
          ]
        },
        "blockA",
        { socketName: "input" }
      )
    ).rejects.toThrowError("block 'blockA' has no output socket with name foo");
    await expect(
      workflowEngine.triggerWorkflowBlock(
        {
          blocks: [
            { id: "blockA", definitionId: MOCK_BLOCK },
            { id: "blockB", definitionId: MOCK_BLOCK }
          ],
          connections: [
            { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "baa" }
          ]
        },
        "blockA",
        { socketName: "input" }
      )
    ).rejects.toThrowError("block 'blockB' has no input socket with name baa");
  });

  it("can't trigger workflows with multiple connections on one output", async () => {
    await expect(
      workflowEngine.triggerWorkflowBlock(
        {
          blocks: [
            { id: "blockA", definitionId: MOCK_BLOCK },
            { id: "blockB", definitionId: MOCK_BLOCK }
          ],
          connections: [
            { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" },
            { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
          ]
        },
        "blockA",
        { socketName: "input" }
      )
    ).rejects.toThrowError("more than one output connection is currently not allowed");
  });

  it("can't trigger workflows with unknown definitions", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: "foo" },
        { id: "blockB", definitionId: "baaa" }
      ],
      connections: [
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" },
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
      ]
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" })
    ).rejects.toThrowError();
  });

  it("can't trigger workflows with unknown blocks", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: "foo" },
        { id: "blockB", definitionId: "baaa" }
      ]
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blöckle", { socketName: "input" })
    ).rejects.toThrowError();
  });

  it("can trigger workflows with blocks that don't return anything", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: MOCK_BLOCK },
        { id: "blockB", definitionId: MOCK_BLOCK_DEADEND }
      ],
      connections: [
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
      ]
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" })
    ).resolves.toMatchObject({ data: undefined });
  });

  it("can trigger workflows with blocks that return crap", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: MOCK_BLOCK },
        { id: "blockB", definitionId: MOCK_BLOCK_STUPIDEND }
      ],
      connections: [
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
      ]
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" })
    ).resolves.toMatchObject({ data: undefined });
  });

  it("can trigger workflows with data", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: MOCK_BLOCK },
        { id: "blockB", definitionId: MOCK_BLOCK }
      ],
      connections: [
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
      ]
    };
    await expect(
      workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input", data: 5 })
    ).resolves.toMatchObject({ data: 7 });
  });

  it("can trigger workflows and read logs", async () => {
    const workflow: Workflow = {
      blocks: [
        { id: "blockA", definitionId: MOCK_BLOCK },
        { id: "blockB", definitionId: MOCK_BLOCK }
      ],
      connections: [
        { sourceBlockId: "blockA", sourceSocketName: "output", targetBlockId: "blockB", targetSocketName: "input" }
      ]
    };
    const executionResult: WorkflowExecutionResult = await workflowEngine.triggerWorkflowBlock(workflow, "blockA", {
      socketName: "input",
      data: 5
    });
    expect(executionResult.logs.length).toEqual(10);
  });

  it("can configure workflow blocks", async () => {
    const workflow: Workflow = {
      blocks: [{ id: "blockA", definitionId: MOCK_BLOCK_CONFIG, configuration: { meins: "hehe" } }]
    };
    const executionResult: WorkflowExecutionResult = await workflowEngine.triggerWorkflowBlock(workflow, "blockA", {
      socketName: "input"
    });
    expect(executionResult.data).toEqual("hehe");
  });

  it("workflow blocks can throw exceptions", async () => {
    const workflow: Workflow = { blocks: [{ id: "blockA", definitionId: MOCK_BLOCK_THROWS }] };
    const logger = new DefaultWorkflowLogger();
    try {
      await workflowEngine.triggerWorkflowBlock(workflow, "blockA", { socketName: "input" }, { logger });
      throw new Error("did not throw exception");
    } catch (e) {
      expect(logger.getLogs()).toMatchObject([
        {
          level: "info",
          socketName: "input",
          text: 'Executing block "mock_block_throws", socket: input',
          extra: JSON.stringify({ blockId: "blockA" }, undefined, 2)
        },
        {
          blockId: "blockA",
          blockName: "mock_block_throws",
          level: "error",
          socketName: "input",
          text: "Uncaught Exception: test"
        }
      ]);
    }
  });
});
