import { SocketConnection, SocketDirection } from "../models/socket";
import {
  Workflow,
  WorkflowExecutionContext,
  WorkflowExecutionResult,
  WorkflowExecutionStatus
} from "../models/workflow";
import {
  BlockExecutionContext,
  BlockExecutionResult,
  WorkflowBlock,
  WorkflowBlockDefinition
} from "../models/workflow-block";
import { DefaultWorkflowLogger, WorkflowLogger } from "./workflow-logger";

export class WorkflowEngine {
  protected blockDefinitions: { [definitionId: string]: WorkflowBlockDefinition } = {};

  public addBlockDefinition(definitionId: string, definition: WorkflowBlockDefinition): void {
    this.blockDefinitions[definitionId] = definition;
  }

  public async getBlockDefinition(definitionId: string): Promise<WorkflowBlockDefinition> {
    const block = this.blockDefinitions[definitionId];
    if (!block) throw new Error("block definition not found by id: " + definitionId);
    return block;
  }

  public getBlockById(workflow: Workflow, blockId: string): WorkflowBlock {
    const workflowBlock = workflow.blocks.find((block) => block.id === blockId);
    if (!workflowBlock) throw new Error("block not found by id: " + blockId);
    return workflowBlock;
  }

  public getSocketConnectionsBySource(
    workflow: Workflow,
    sourceBlockId: string,
    sourceSocketName: string
  ): SocketConnection[] {
    return (
      workflow.connections &&
      workflow.connections.filter(
        (connection: SocketConnection) =>
          connection.sourceBlockId === sourceBlockId && connection.sourceSocketName === sourceSocketName
      )
    );
  }

  public getLogger(): WorkflowLogger {
    return new DefaultWorkflowLogger();
  }

  // eslint-disable-next-line max-lines-per-function
  public async triggerWorkflowBlock(
    workflow: Workflow,
    blockId: string,
    blockExecutionContext: BlockExecutionContext,
    workflowExecutionContext?: WorkflowExecutionContext
  ): Promise<WorkflowExecutionResult> {
    if (workflowExecutionContext === undefined) workflowExecutionContext = { logger: this.getLogger() };

    const block: WorkflowBlock = this.getBlockById(workflow, blockId);
    const blockDefinition: WorkflowBlockDefinition = await this.getBlockDefinition(block.definitionId);
    workflowExecutionContext.logger.setContext(blockId, block.definitionId, blockExecutionContext.socketName);

    await workflowExecutionContext.logger.info(
      `Executing block "${block.definitionId}", socket: ${blockExecutionContext.socketName}`,
      JSON.stringify(
        {
          blockId,
          data: blockExecutionContext.data,
          configuration: block.configuration
        },
        undefined,
        2
      ).substring(0, 2048)
    );

    // check sockets and connections
    if (
      blockDefinition
        .getSockets()
        .find(
          (socket) => socket.name === blockExecutionContext.socketName && socket.direction === SocketDirection.IN
        ) === undefined
    )
      throw new Error("block '" + blockId + "' has no input socket with name " + blockExecutionContext.socketName);

    if (workflow.connections instanceof Array) {
      for (const connection of workflow.connections) {
        // check start block
        const startBlock: WorkflowBlock = this.getBlockById(workflow, connection.sourceBlockId);
        const startBlockDefinition: WorkflowBlockDefinition = await this.getBlockDefinition(startBlock.definitionId);
        if (
          startBlockDefinition
            .getSockets()
            .find(
              (socket) => socket.name === connection.sourceSocketName && socket.direction === SocketDirection.OUT
            ) === undefined
        )
          throw new Error(
            "block '" + connection.sourceBlockId + "' has no output socket with name " + connection.sourceSocketName
          );
        // check end block
        const endBlock: WorkflowBlock = this.getBlockById(workflow, connection.targetBlockId);
        const endBlockDefinition: WorkflowBlockDefinition = await this.getBlockDefinition(endBlock.definitionId);
        if (
          endBlockDefinition
            .getSockets()
            .find(
              (socket) => socket.name === connection.targetSocketName && socket.direction === SocketDirection.IN
            ) === undefined
        )
          throw new Error(
            "block '" + connection.targetBlockId + "' has no input socket with name " + connection.targetSocketName
          );
      }
    }

    let executionResult: BlockExecutionResult | void;
    try {
      // execute workflow block
      blockExecutionContext.configuration = block.configuration;
      executionResult = await blockDefinition.executeBlock(blockExecutionContext, workflowExecutionContext);
    } catch (e) {
      await workflowExecutionContext.logger.error("Uncaught Exception: " + e.message);
      return {
        logs: workflowExecutionContext.logger.getLogs(),
        data: undefined,
        status: WorkflowExecutionStatus.FINISHED_WITH_ERRORS
      };
    }
    if (executionResult) {
      if (executionResult.socketName) {
        const connections: SocketConnection[] = this.getSocketConnectionsBySource(
          workflow,
          blockId,
          executionResult.socketName
        );
        if (connections instanceof Array && connections.length > 0) {
          if (connections.length > 1) throw new Error("more than one output connection is currently not allowed");
          const newBlockExecutionContext: BlockExecutionContext = {
            socketName: connections[0].targetSocketName,
            data: executionResult.data
          };
          return this.triggerWorkflowBlock(
            workflow,
            connections[0].targetBlockId,
            newBlockExecutionContext,
            workflowExecutionContext
          );
        } else {
          return {
            data: executionResult.data,
            logs: workflowExecutionContext.logger.getLogs(),
            status: WorkflowExecutionStatus.FINISHED_SUCCESSFULLY
          };
        }
      } else {
        return {
          data: executionResult.data,
          logs: workflowExecutionContext.logger.getLogs(),
          status: WorkflowExecutionStatus.FINISHED_SUCCESSFULLY
        };
      }
    }
    return {
      logs: workflowExecutionContext.logger.getLogs(),
      data: undefined,
      status: WorkflowExecutionStatus.FINISHED_SUCCESSFULLY
    };
  }
}
